# Script to plot MAGMA performance data. See help below and readme.txt.
# @author Mark Gates

from __future__ import print_function

import matplotlib.pyplot as pp
import matplotlib
import numpy
from numpy import isnan, zeros, where, linspace

import util


# ----------------------------------------------------------------------
# import versions
import v150_cuda70_k40c
import v160_cuda70_k40c
import v161_cuda70_k40c
import v162_cuda70_k40c  # same as v161 except sy/heevd
import v170_cuda70_k40c
import v200_cuda70_k40c

versions = [
	v150_cuda70_k40c,
	v160_cuda70_k40c,
	v161_cuda70_k40c,
	v162_cuda70_k40c,
	v170_cuda70_k40c,
	v200_cuda70_k40c,
]

# add local if it exists
try:
	import local
	versions.append( local )
	local.version = 'local'
except ImportError:
	pass

versions = numpy.array( versions )


# ----------------------------------------------------------------------
# setup colors for versions

# get nice distribution of colors from purple (old versions) to red (new versions)
#x = linspace( 0, 1, len(versions) )
#cool = matplotlib.cm.get_cmap('cool')
#colors = cool(x)

colors = [
 'r',
 'orange',
 'c',
 'b',
 'g',
 'm',
 'indigo',
 'SlateBlue',
 'LimeGreen',
]

# set colors, with latest being red (first)
for i in xrange( len(versions) ):
	#versions[i].color = colors[i,:]
	versions[i].color = colors[ len(versions) - 1 - i ]
# end


# ----------------------------------------------------------------------
# defaults
pp.rcParams['axes.axisbelow']   = True
pp.rcParams['figure.facecolor'] = 'white'
pp.rcParams['font.size']        = 10
pp.rcParams['legend.fontsize']  = 9

g_figsize  = [9, 7]
g_figsize2 = [6, 4]


# ----------------------------------------------------------------------
if ( not locals().has_key('g_subplots')):
	g_subplots = True
if ( not locals().has_key('g_figure')):
	g_figure   = None
if ( not locals().has_key('g_log')):
	g_log      = True
if ( not locals().has_key('g_save')):
	g_save     = False


# --------------------
def figure( fig ):
	global g_figure
	g_figure = fig
	if ( g_subplots ):
		pp.figure( fig )
# end


# --------------------
def clf( rows=0, cols=0 ):
	if ( g_subplots or rows == 0 or cols == 0 ):
		pp.clf()
	else:
		for index in range( 1, rows*cols+1 ):
			subplot( rows, cols, index )
			pp.clf()
# end


# --------------------
def subplot( rows, cols, index ):
	if ( g_subplots ):
		pp.subplot( rows, cols, index )
	else:
		pp.figure( g_figure*10 + index )
# end


# --------------------
def resize( size, rows=0, cols=0 ):
	if ( g_subplots or rows == 0 or cols == 0 ):
		util.resize( size )
	else:
		size2 = [ size[0]/cols, size[1]/rows ]
		for index in range( 1, rows*cols+1 ):
			subplot( rows, cols, index )
			util.resize( g_figsize2 )
# end


# --------------------
def savefig( name, rows=0, cols=0 ):
	if ( g_save ):
		if ( g_subplots or rows == 0 or cols == 0 ):
			print( 'saving', name )
			pp.savefig( name )
		else:
			prec = ['s', 'd', 'c', 'z']
			for index in range( 1, rows*cols+1 ):
				pname = prec[index-1] + name
				print( 'saving', pname )
				subplot( rows, cols, index )
				pp.savefig( pname )
# end


# --------------------
def set_title( versions, title ):
	v0 = versions[0]
	t = None
	for v in versions[1:]:
		if (v.cpu != v0.cpu) or (v.device != v0.device):
			t = 'various CPU & GPU'
			break
	if ( t == None ):
		t = 'GPU: ' + v0.device + ', CPU: ' + v0.cpu
	if ( title ):
		t = title + '\n' + t
	pp.title( t, fontsize=9 )
# end


# --------------------
def set_xticks():
	if ( g_log ):
		pp.xlabel( r'matrix dimension (log scale)' )
		xticks = [ 10, 100, 1000, 10000, 20000 ]
		xtickl = [ 10, 100, '1k', '10k', '20k' ]
	else:
		pp.xlabel( r'matrix dimension' )
		xticks = range( 0, 20001, 4000 )
		xtickl = xticks
	pp.xticks( xticks, xtickl )
	pp.xlim( 9, 20000 )
	pp.grid( True )
# end


# --------------------
# returns list of versions -- either original versions, or if it is single module, stick it in a list
module = type(numpy)  # why doesn't python know "module"?
def get_versions( versions ):
	if ( type(versions) == module ):
		return [versions]
	else:
		return versions
# end


# ----------------------------------------------------------------------
# column indices
getrf_m         = 0
getrf_n         = 1
getrf_cpu_flops = 2
getrf_cpu_time  = 3
getrf_gpu_flops = 4
getrf_gpu_time  = 5
getrf_error     = 6

potrf_n         = 0
potrf_cpu_flops = 1
potrf_cpu_time  = 2
potrf_gpu_flops = 3
potrf_gpu_time  = 4
potrf_error     = 5

geqrf_m         = 0
geqrf_n         = 1
geqrf_cpu_flops = 2
geqrf_cpu_time  = 3
geqrf_gpu_flops = 4
geqrf_gpu_time  = 5
geqrf_error     = 6

geev_n          = 0
geev_cpu_time   = 1
geev_gpu_time   = 2
geev_error      = 3

syev_n          = 0
syev_cpu_time   = 1
syev_gpu_time   = 2
syev_error      = 3

# note: testing_.gesvd outputs jobu & jobv; we throw out jobv in python data file
# to match testing_.gesdd which has only job.
svd_job         = 0
svd_m           = 1
svd_n           = 2
svd_cpu_time    = 3
svd_gpu_time    = 4
svd_error       = 5

symv_n             = 0
symv_gpu_flops     = 1
symv_gpu_time      = 2
symv_atomics_flops = 3
symv_atomics_time  = 4
symv_cublas_flops  = 5
symv_cublas_time   = 6
symv_cpu_flops     = 7
symv_cpu_time      = 8
symv_gpu_error     = 9
symv_atomics_error = 10
symv_cublas_error  = 11


# ----------------------------------------------------------------------
def str_none( s ):
	return str(s) if s is not None else ''


def plot_getrf_data( data, style='.-', color='y', label=None, idx=getrf_gpu_flops ):
	if ( g_log ):
		pp.semilogx( data[:,getrf_m], data[:,idx], style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     data[:,getrf_m], data[:,idx], style, color=color, lw=1.5, label=label )
# end

def plot_getrf_labels( versions, title=None, legend=True ):
	set_title( versions, 'LU factorization ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	pp.ylabel( r'Gflop/s  $\frac{2}{3} n^3 / t$' )
	set_xticks()
# end

def plot_getrf( versions, cpu=True, gpu=True, lapack=True ):
	print( 'plotting getrf' )
	versions = get_versions( versions )
	figure( 1 )
	clf( 2, 2 )
	
	for v in versions:
		# ----- cpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgetrf') and cpu:
			plot_getrf_data( v.sgetrf,     '.-', color=v.color, label=v.version+' sgetrf'     )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgetrf') and cpu:
			plot_getrf_data( v.dgetrf,     '.-', color=v.color, label=v.version+' dgetrf'     )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgetrf') and cpu:
			plot_getrf_data( v.cgetrf,     '.-', color=v.color, label=v.version+' cgetrf'     )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgetrf') and cpu:
			plot_getrf_data( v.zgetrf,     '.-', color=v.color, label=v.version+' zgetrf'     )
		
		# ----- gpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgetrf_gpu') and gpu:
			plot_getrf_data( v.sgetrf_gpu, 'x-', color=v.color, label=v.version+' sgetrf_gpu' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgetrf_gpu') and gpu:
			plot_getrf_data( v.dgetrf_gpu, 'x-', color=v.color, label=v.version+' dgetrf_gpu' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgetrf_gpu') and gpu:
			plot_getrf_data( v.cgetrf_gpu, 'x-', color=v.color, label=v.version+' cgetrf_gpu' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgetrf_gpu') and gpu:
			plot_getrf_data( v.zgetrf_gpu, 'x-', color=v.color, label=v.version+' zgetrf_gpu' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgetrf') and not isnan( v.sgetrf[0,getrf_cpu_flops] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgetrf'):
				plot_getrf_data( v.sgetrf, 'k+-', color='k', label='MKL sgetrf', idx=getrf_cpu_flops )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgetrf'):
				plot_getrf_data( v.dgetrf, 'k+-', color='k', label='MKL dgetrf', idx=getrf_cpu_flops )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgetrf'):
				plot_getrf_data( v.cgetrf, 'k+-', color='k', label='MKL cgetrf', idx=getrf_cpu_flops )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgetrf'):
				plot_getrf_data( v.zgetrf, 'k+-', color='k', label='MKL zgetrf', idx=getrf_cpu_flops )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_getrf_labels( versions, 'sgetrf', legend= True )
	
	subplot( 2, 2, 2 )
	plot_getrf_labels( versions, 'dgetrf', legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_getrf_labels( versions, 'cgetrf', legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_getrf_labels( versions, 'zgetrf', legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'getrf.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_potrf_data( data, style='.-', color='y', label=None, idx=potrf_gpu_flops ):
	if ( g_log ):
		pp.semilogx( data[:,potrf_n], data[:,idx], style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     data[:,potrf_n], data[:,idx], style, color=color, lw=1.5, label=label )
# end

def plot_potrf_labels( versions, title=None, legend=True ):
	set_title( versions, 'Cholesky factorization ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	pp.ylabel( r'Gflop/s  $\frac{1}{3} n^3 / t$' )
	set_xticks()
# end

def plot_potrf( versions, cpu=True, gpu=True, lapack=True ):
	print( 'plotting potrf' )
	versions = get_versions( versions )
	figure( 2 )
	clf( 2, 2 )
	
	for v in versions:
		# ----- cpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('spotrf') and cpu:
			plot_potrf_data( v.spotrf,     '.-', color=v.color, label=v.version+' spotrf'     )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dpotrf') and cpu:
			plot_potrf_data( v.dpotrf,     '.-', color=v.color, label=v.version+' dpotrf'     )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cpotrf') and cpu:
			plot_potrf_data( v.cpotrf,     '.-', color=v.color, label=v.version+' cpotrf'     )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zpotrf') and cpu:
			plot_potrf_data( v.zpotrf,     '.-', color=v.color, label=v.version+' zpotrf'     )
		
		# ----- gpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('spotrf_gpu') and gpu:
			plot_potrf_data( v.spotrf_gpu, 'x-', color=v.color, label=v.version+' spotrf_gpu' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dpotrf_gpu') and gpu:
			plot_potrf_data( v.dpotrf_gpu, 'x-', color=v.color, label=v.version+' dpotrf_gpu' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cpotrf_gpu') and gpu:
			plot_potrf_data( v.cpotrf_gpu, 'x-', color=v.color, label=v.version+' cpotrf_gpu' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zpotrf_gpu') and gpu:
			plot_potrf_data( v.zpotrf_gpu, 'x-', color=v.color, label=v.version+' zpotrf_gpu' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('spotrf') and not isnan( v.spotrf[0,potrf_cpu_flops] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('spotrf'):
				plot_potrf_data( v.spotrf, 'k+-', color='k', label='MKL spotrf', idx=potrf_cpu_flops )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dpotrf'):
				plot_potrf_data( v.dpotrf, 'k+-', color='k', label='MKL dpotrf', idx=potrf_cpu_flops )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cpotrf'):
				plot_potrf_data( v.cpotrf, 'k+-', color='k', label='MKL cpotrf', idx=potrf_cpu_flops )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zpotrf'):
				plot_potrf_data( v.zpotrf, 'k+-', color='k', label='MKL zpotrf', idx=potrf_cpu_flops )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_potrf_labels( versions, 'spotrf', legend= True )
	
	subplot( 2, 2, 2 )
	plot_potrf_labels( versions, 'dpotrf', legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_potrf_labels( versions, 'cpotrf', legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_potrf_labels( versions, 'zpotrf', legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'potrf.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_geqrf_data( data, style='.-', color='y', label=None, idx=geqrf_gpu_flops ):
	if ( g_log ):
		pp.semilogx( data[:,geqrf_m], data[:,idx], style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     data[:,geqrf_m], data[:,idx], style, color=color, lw=1.5, label=label )
# end

def plot_geqrf_labels( versions, title=None, legend=True ):
	set_title( versions, 'QR factorization ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	pp.ylabel( r'Gflop/s  $\frac{4}{3} n^3 / t$' )
	set_xticks()
# end

def plot_geqrf( versions, cpu=True, gpu=True, lapack=True ):
	print( 'plotting geqrf' )
	versions = get_versions( versions )
	figure( 3 )
	clf( 2, 2 )
	
	for v in versions:
		# ----- cpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgeqrf') and cpu:
			plot_geqrf_data( v.sgeqrf,     '.-', color=v.color, label=v.version+' sgeqrf'     )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgeqrf') and cpu:
			plot_geqrf_data( v.dgeqrf,     '.-', color=v.color, label=v.version+' dgeqrf'     )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgeqrf') and cpu:
			plot_geqrf_data( v.cgeqrf,     '.-', color=v.color, label=v.version+' cgeqrf'     )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgeqrf') and cpu:
			plot_geqrf_data( v.zgeqrf,     '.-', color=v.color, label=v.version+' zgeqrf'     )
		
		# ----- gpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgeqrf_gpu') and gpu:
			plot_geqrf_data( v.sgeqrf_gpu, 'x-', color=v.color, label=v.version+' sgeqrf_gpu' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgeqrf_gpu') and gpu:
			plot_geqrf_data( v.dgeqrf_gpu, 'x-', color=v.color, label=v.version+' dgeqrf_gpu' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgeqrf_gpu') and gpu:
			plot_geqrf_data( v.cgeqrf_gpu, 'x-', color=v.color, label=v.version+' cgeqrf_gpu' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgeqrf_gpu') and gpu:
			plot_geqrf_data( v.zgeqrf_gpu, 'x-', color=v.color, label=v.version+' zgeqrf_gpu' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgeqrf') and not isnan( v.sgeqrf[0,geqrf_cpu_flops] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgeqrf'):
				plot_geqrf_data( v.sgeqrf, 'k+-', color='k', label='MKL sgeqrf', idx=geqrf_cpu_flops )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgeqrf'):
				plot_geqrf_data( v.dgeqrf, 'k+-', color='k', label='MKL dgeqrf', idx=geqrf_cpu_flops )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgeqrf'):
				plot_geqrf_data( v.cgeqrf, 'k+-', color='k', label='MKL cgeqrf', idx=geqrf_cpu_flops )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgeqrf'):
				plot_geqrf_data( v.zgeqrf, 'k+-', color='k', label='MKL zgeqrf', idx=geqrf_cpu_flops )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_geqrf_labels( versions, 'sgeqrf', legend= True )
	
	subplot( 2, 2, 2 )
	plot_geqrf_labels( versions, 'dgeqrf', legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_geqrf_labels( versions, 'cgeqrf', legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_geqrf_labels( versions, 'zgeqrf', legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'geqrf.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_geev_data( data, vec, style='.-', color='y', label=None, idx=geev_gpu_time ):
	n = data[:,geev_n]
	t = data[:,idx]
	if ( vec ):
		# 10/3 Hess + 10/3 QR iter + 4/3 generate Q + 4/3 right vectors
		gflop = 1e-9 * 28/3. * n**3
	else:
		# 10/3 Hess +  4/3 QR iter
		gflop = 1e-9 * 14/3. * n**3
	if ( g_log ):
		pp.semilogx( n, gflop/t, style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     n, gflop/t, style, color=color, lw=1.5, label=label )
# end

def plot_geev_labels( versions, title, vec, legend=True ):
	set_title( versions, 'Non-symmetric eigenvalues, ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	if ( vec ):
		pp.ylabel( r'Gflop/s   $\frac{28}{3} n^3 / t$' )
	else:
		pp.ylabel( r'Gflop/s   $\frac{14}{3} n^3 / t$' )
	#pp.ylabel( 'time (sec)' )
	set_xticks()
# end

def plot_geev( versions, lapack=True ):
	print( 'plotting geev' )
	versions = get_versions( versions )
	figure( 4 )
	clf( 2, 2 )
	
	for v in versions:
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgeev_RN'):
			plot_geev_data(  v.sgeev_RN, False, '.-', color=v.color, label=v.version+' sgeev' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgeev_RN'):
			plot_geev_data(  v.dgeev_RN, False, '.-', color=v.color, label=v.version+' dgeev' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgeev_RN'):
			plot_geev_data(  v.cgeev_RN, False, '.-', color=v.color, label=v.version+' cgeev' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgeev_RN'):
			plot_geev_data(  v.zgeev_RN, False, '.-', color=v.color, label=v.version+' zgeev' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgeev_RN') and not isnan( v.sgeev_RN[0,geev_cpu_time] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgeev_RN'):
				plot_geev_data(  v.sgeev_RN, False, 'k+-', color='k', label='MKL sgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgeev_RN'):
				plot_geev_data(  v.dgeev_RN, False, 'k+-', color='k', label='MKL dgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgeev_RN'):
				plot_geev_data(  v.cgeev_RN, False, 'k+-', color='k', label='MKL cgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgeev_RN'):
				plot_geev_data(  v.zgeev_RN, False, 'k+-', color='k', label='MKL zgeev', idx=geev_cpu_time )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_geev_labels( versions, 'sgeev, no vectors', vec=False, legend= True )
	
	subplot( 2, 2, 2 )
	plot_geev_labels( versions, 'dgeev, no vectors', vec=False, legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_geev_labels( versions, 'cgeev, no vectors', vec=False, legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_geev_labels( versions, 'zgeev, no vectors', vec=False, legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'geev_rn.pdf', 2, 2 )
	
	# --------------------
	figure( 5 )
	clf( 2, 2 )
	
	for v in versions:
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgeev_RV'):
			plot_geev_data(  v.sgeev_RV, True, '.-', color=v.color, label=v.version+' sgeev' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgeev_RV'):
			plot_geev_data(  v.dgeev_RV, True, '.-', color=v.color, label=v.version+' dgeev' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgeev_RV'):
			plot_geev_data(  v.cgeev_RV, True, '.-', color=v.color, label=v.version+' cgeev' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgeev_RV'):
			plot_geev_data(  v.zgeev_RV, True, '.-', color=v.color, label=v.version+' zgeev' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgeev_RV') and not isnan( v.sgeev_RV[0,geev_cpu_time] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgeev_RV'):
				plot_geev_data(  v.sgeev_RV, True, 'k+-', color='k', label='MKL sgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgeev_RV'):
				plot_geev_data(  v.dgeev_RV, True, 'k+-', color='k', label='MKL dgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgeev_RV'):
				plot_geev_data(  v.cgeev_RV, True, 'k+-', color='k', label='MKL cgeev', idx=geev_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgeev_RV'):
				plot_geev_data(  v.zgeev_RV, True, 'k+-', color='k', label='MKL zgeev', idx=geev_cpu_time )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_geev_labels( versions, 'sgeev, with right vectors', vec=True, legend= True )
	
	subplot( 2, 2, 2 )
	plot_geev_labels( versions, 'dgeev, with right vectors', vec=True, legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_geev_labels( versions, 'cgeev, with right vectors', vec=True, legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_geev_labels( versions, 'zgeev, with right vectors', vec=True, legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'geev_rv.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_syev_data( data, vec, style='.-', color='y', label=None, idx=syev_gpu_time ):
	n = data[:,syev_n]
	t = data[:,idx]  # idx is syev_cpu_time or syev_gpu_time
	if ( vec ):
		# 4/3 sytrd + 4/3 solve + 2 mqr
		gflop = 1e-9 * 14/3. * n**3
	else:
		gflop = 1e-9 * 4/3. * n**3
	
	if ( g_log ):
		pp.semilogx( n, gflop/t, style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     n, gflop/t, style, color=color, lw=1.5, label=label )
# end

def plot_syev_labels( versions, title, vec, legend=True ):
	set_title( versions, 'Symmetric eigenvalues, ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	if ( vec ):
		pp.ylabel( r'Gflop/s   $\frac{14}{3} n^3 / t$' )
	else:
		pp.ylabel( r'Gflop/s   $\frac{4}{3} n^3 / t$' )
	#pp.ylabel( 'time (sec)' )
	set_xticks()
# end

def plot_syev( versions, cpu=True, gpu=True, bulge=True, lapack=True ):
	print( 'plotting syev' )
	versions = get_versions( versions )
	figure( 6 )
	clf( 2, 2 )
	
	for v in versions:
		# ---- bulge
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevdx_2stage_JN') and bulge:
			plot_syev_data(  v.ssyevdx_2stage_JN, False, 's-', color=v.color, label=v.version+' ssyevdx_2stage' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevdx_2stage_JN') and bulge:
			plot_syev_data(  v.dsyevdx_2stage_JN, False, 's-', color=v.color, label=v.version+' dsyevdx_2stage' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevdx_2stage_JN') and bulge:
			plot_syev_data(  v.cheevdx_2stage_JN, False, 's-', color=v.color, label=v.version+' cheevdx_2stage' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevdx_2stage_JN') and bulge:
			plot_syev_data(  v.zheevdx_2stage_JN, False, 's-', color=v.color, label=v.version+' zheevdx_2stage' )
		
		# ----- cpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevd_JN') and cpu:
			plot_syev_data(  v.ssyevd_JN,     False, '.-', color=v.color, label=v.version+' ssyevd'     )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevd_JN') and cpu:
			plot_syev_data(  v.dsyevd_JN,     False, '.-', color=v.color, label=v.version+' dsyevd'     )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevd_JN') and cpu:
			plot_syev_data(  v.cheevd_JN,     False, '.-', color=v.color, label=v.version+' cheevd'     )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevd_JN') and cpu:
			plot_syev_data(  v.zheevd_JN,     False, '.-', color=v.color, label=v.version+' zheevd'     )
		
		# ----- gpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevd_gpu_JN') and gpu:
			plot_syev_data(  v.ssyevd_gpu_JN, False, 'x-', color=v.color, label=v.version+' ssyevd_gpu' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevd_gpu_JN') and gpu:
			plot_syev_data(  v.dsyevd_gpu_JN, False, 'x-', color=v.color, label=v.version+' dsyevd_gpu' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevd_gpu_JN') and gpu:
			plot_syev_data(  v.cheevd_gpu_JN, False, 'x-', color=v.color, label=v.version+' cheevd_gpu' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevd_gpu_JN') and gpu:
			plot_syev_data(  v.zheevd_gpu_JN, False, 'x-', color=v.color, label=v.version+' zheevd_gpu' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('ssyevd_JN') and not isnan( v.ssyevd_JN[0,syev_cpu_time] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('ssyevd_JN'):
				plot_syev_data(  v.ssyevd_JN,     False, 'k+-', color='k', label='MKL ssyevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dsyevd_JN'):
				plot_syev_data(  v.dsyevd_JN,     False, 'k+-', color='k', label='MKL dsyevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cheevd_JN'):
				plot_syev_data(  v.cheevd_JN,     False, 'k+-', color='k', label='MKL cheevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zheevd_JN'):
				plot_syev_data(  v.zheevd_JN,     False, 'k+-', color='k', label='MKL zheevd', idx=syev_cpu_time )
			break
	# end
	
	subplot( 2, 2, 1 )
	plot_syev_labels( versions, 'ssyevd, no vectors', vec=False, legend= True )
	
	subplot( 2, 2, 2 )
	plot_syev_labels( versions, 'dsyevd, no vectors', vec=False, legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_syev_labels( versions, 'cheevd, no vectors', vec=False, legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_syev_labels( versions, 'zheevd, no vectors', vec=False, legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'syev_jn.pdf', 2, 2 )
	
	# --------------------
	figure( 7 )
	clf( 2, 2 )
	
	for v in versions:
		# ----- bulge
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevdx_2stage_JV') and bulge:
			plot_syev_data(  v.ssyevdx_2stage_JV, True, 's-', color=v.color, label=v.version+' ssyevdx_2stage' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevdx_2stage_JV') and bulge:
			plot_syev_data(  v.dsyevdx_2stage_JV, True, 's-', color=v.color, label=v.version+' dsyevdx_2stage' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevdx_2stage_JV') and bulge:
			plot_syev_data(  v.cheevdx_2stage_JV, True, 's-', color=v.color, label=v.version+' cheevdx_2stage' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevdx_2stage_JV') and bulge:
			plot_syev_data(  v.zheevdx_2stage_JV, True, 's-', color=v.color, label=v.version+' zheevdx_2stage' )
		
		# ----- cpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevd_JV') and cpu:
			plot_syev_data(  v.ssyevd_JV,     True, '.-', color=v.color, label=v.version+' ssyevd'     )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevd_JV') and cpu:
			plot_syev_data(  v.dsyevd_JV,     True, '.-', color=v.color, label=v.version+' dsyevd'     )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevd_JV') and cpu:
			plot_syev_data(  v.cheevd_JV,     True, '.-', color=v.color, label=v.version+' cheevd'     )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevd_JV') and cpu:
			plot_syev_data(  v.zheevd_JV,     True, '.-', color=v.color, label=v.version+' zheevd'     )
		
		# ----- gpu
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssyevd_gpu_JV') and gpu:
			plot_syev_data(  v.ssyevd_gpu_JV, True, 'x-', color=v.color, label=v.version+' ssyevd_gpu' )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsyevd_gpu_JV') and gpu:
			plot_syev_data(  v.dsyevd_gpu_JV, True, 'x-', color=v.color, label=v.version+' dsyevd_gpu' )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cheevd_gpu_JV') and gpu:
			plot_syev_data(  v.cheevd_gpu_JV, True, 'x-', color=v.color, label=v.version+' cheevd_gpu' )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zheevd_gpu_JV') and gpu:
			plot_syev_data(  v.zheevd_gpu_JV, True, 'x-', color=v.color, label=v.version+' zheevd_gpu' )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('ssyevd_JV') and not isnan( v.ssyevd_JV[0,syev_cpu_time] )):
			print( 'found LAPACK in', v.version )
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('ssyevd_JV'):
				plot_syev_data(  v.ssyevd_JV,     True, 'k+-', color='k', label='MKL ssyevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dsyevd_JV'):
				plot_syev_data(  v.dsyevd_JV,     True, 'k+-', color='k', label='MKL dsyevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cheevd_JV'):
				plot_syev_data(  v.cheevd_JV,     True, 'k+-', color='k', label='MKL cheevd', idx=syev_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zheevd_JV'):
				plot_syev_data(  v.zheevd_JV,     True, 'k+-', color='k', label='MKL zheevd', idx=syev_cpu_time )
			break
	# end
	
	for i in xrange( 1, 5 ):
		subplot( 2, 2, 1 )
		plot_syev_labels( versions, 'ssyevd, with vectors', vec=True, legend= True )
		
		subplot( 2, 2, 2 )
		plot_syev_labels( versions, 'dsyevd, with vectors', vec=True, legend= not g_subplots )
		
		subplot( 2, 2, 3 )
		plot_syev_labels( versions, 'cheevd, with vectors', vec=True, legend= not g_subplots )
		
		subplot( 2, 2, 4 )
		plot_syev_labels( versions, 'zheevd, with vectors', vec=True, legend= not g_subplots )
	# end
	resize( g_figsize, 2, 2 )
	savefig( 'syev_jv.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_gesvd_data( data, vec, style='.-', color='y', label=None, ratio=1, idx=svd_gpu_time ):
	(ii,) = where( data[:,svd_m] == ratio*data[:,svd_n] )
	mn = zeros(( len(ii), 2 ))
	mn[:,0] = data[ii,svd_m]
	mn[:,1] = data[ii,svd_n]
	M = mn.max( axis=1 )  # M = max(m,n)
	N = mn.min( axis=1 )  # N = min(m,n)
	
	if ( vec ):
		# for D&C some vectors (no initial QR)
		gflop = 1e-9 * (8*M*N**2 + (4/3.)*N**3)
	else:
		gflop = 1e-9 * (4*M*N**2 - (4/3.)*N**3)
	t = data[ii,idx]
	
	if ( g_log ):
		pp.semilogx( N, gflop/t, style, color=color, lw=1.5, label=label )
	else:
		pp.plot(     N, gflop/t, style, color=color, lw=1.5, label=label )
# end

def plot_gesvd_labels( versions, title, vec, square, legend=True ):
	set_title( versions, 'SVD, ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	if ( vec ):
		if ( square ):
			pp.ylabel( r'Gflop/s   $9n^3/t$' )
		else:
			pp.ylabel( r'Gflop/s   $(8mn^2 + \frac{4}{3}n^3)/t$' )
	else:
		if ( square ):
			pp.ylabel( r'Gflop/s   $\frac{8}{3} n^3 / t$' )
		else:
			pp.ylabel( r'Gflop/s   $(4mn^2 - \frac{4}{3} n^3) / t$' )
	#pp.ylabel( 'time (sec)' )
	set_xticks()
# end

def plot_gesvd( versions, ratio=1, lapack=True, svd=True, sdd=True ):
	print( 'plotting gesvd' )
	versions = get_versions( versions )
	figure( 8 )
	clf( 2, 2 )
	
	for v in versions:
		# for no vectors, gesvd == gesdd
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgesdd_UN') and sdd:
			plot_gesvd_data( v.sgesdd_UN, False, 'x--', color=v.color, label=v.version+' sgesdd', ratio=ratio )
		if v.__dict__.has_key('sgesvd_UN') and svd:
			plot_gesvd_data( v.sgesvd_UN, False, '.-',  color=v.color, label=v.version+' sgesvd', ratio=ratio )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgesdd_UN') and sdd:
			plot_gesvd_data( v.dgesdd_UN, False, 'x--', color=v.color, label=v.version+' dgesdd', ratio=ratio )
		if v.__dict__.has_key('dgesvd_UN') and svd:
			plot_gesvd_data( v.dgesvd_UN, False, '.-',  color=v.color, label=v.version+' dgesvd', ratio=ratio )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgesdd_UN') and sdd:
			plot_gesvd_data( v.cgesdd_UN, False, 'x--', color=v.color, label=v.version+' cgesdd', ratio=ratio )
		if v.__dict__.has_key('cgesvd_UN') and svd:
			plot_gesvd_data( v.cgesvd_UN, False, '.-',  color=v.color, label=v.version+' cgesvd', ratio=ratio )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgesdd_UN') and sdd:
			plot_gesvd_data( v.zgesdd_UN, False, 'x--', color=v.color, label=v.version+' zgesdd', ratio=ratio )
		if v.__dict__.has_key('zgesvd_UN') and svd:
			plot_gesvd_data( v.zgesvd_UN, False, '.-',  color=v.color, label=v.version+' zgesvd', ratio=ratio )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgesdd_UN') and not isnan( v.sgesdd_UN[0,svd_cpu_time] )):
			print( 'found LAPACK in', v.version )
			
			# for no vectors, gesvd == gesdd
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgesdd_UN') and sdd:
				plot_gesvd_data( v.sgesdd_UN, False, 'x--', color='k', label='MKL sgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('sgesvd_UN') and svd:
				plot_gesvd_data( v.sgesvd_UN, False, '.-',  color='k', label='MKL sgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgesdd_UN') and sdd:
				plot_gesvd_data( v.dgesdd_UN, False, 'x--', color='k', label='MKL dgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('dgesvd_UN') and svd:
				plot_gesvd_data( v.dgesvd_UN, False, '.-',  color='k', label='MKL dgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgesdd_UN') and sdd:
				plot_gesvd_data( v.cgesdd_UN, False, 'x--', color='k', label='MKL cgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('cgesvd_UN') and svd:
				plot_gesvd_data( v.cgesvd_UN, False, '.-',  color='k', label='MKL cgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgesdd_UN') and sdd:
				plot_gesvd_data( v.zgesdd_UN, False, 'x--', color='k', label='MKL zgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('zgesvd_UN') and svd:
				plot_gesvd_data( v.zgesvd_UN, False, '.-',  color='k', label='MKL zgesvd', ratio=ratio, idx=svd_cpu_time )
			break
	# end
	
	m = ratio if (ratio >= 1) else 1
	n = 1     if (ratio >= 1) else 1/ratio
	subplot( 2, 2, 1 )
	plot_gesvd_labels( versions, 'sgesvd, no vectors, M:N ratio %.3g:%.3g' % (m,n), vec=False, square=(ratio == 1), legend= True )
	
	subplot( 2, 2, 2 )
	plot_gesvd_labels( versions, 'dgesvd, no vectors, M:N ratio %.3g:%.3g' % (m,n), vec=False, square=(ratio == 1), legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_gesvd_labels( versions, 'cgesvd, no vectors, M:N ratio %.3g:%.3g' % (m,n), vec=False, square=(ratio == 1), legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_gesvd_labels( versions, 'zgesvd, no vectors, M:N ratio %.3g:%.3g' % (m,n), vec=False, square=(ratio == 1), legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'gesvd_jn.pdf', 2, 2 )
	
	# --------------------
	figure( 9 )
	clf( 2, 2 )
	
	for v in versions:
		# for vectors, gesdd > gesvd performance
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('sgesdd_US') and sdd:
			plot_gesvd_data( v.sgesdd_US, True, 'x--', color=v.color, label=v.version+' sgesdd', ratio=ratio )
		if v.__dict__.has_key('sgesvd_US') and svd:
			plot_gesvd_data( v.sgesvd_US, True, '.-',  color=v.color, label=v.version+' sgesvd', ratio=ratio )
			
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dgesdd_US') and sdd:
			plot_gesvd_data( v.dgesdd_US, True, 'x--', color=v.color, label=v.version+' dgesdd', ratio=ratio )
		if v.__dict__.has_key('dgesvd_US') and svd:
			plot_gesvd_data( v.dgesvd_US, True, '.-',  color=v.color, label=v.version+' dgesvd', ratio=ratio )
			
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('cgesdd_US') and sdd:
			plot_gesvd_data( v.cgesdd_US, True, 'x--', color=v.color, label=v.version+' cgesdd', ratio=ratio )
		if v.__dict__.has_key('cgesvd_US') and svd:
			plot_gesvd_data( v.cgesvd_US, True, '.-',  color=v.color, label=v.version+' cgesvd', ratio=ratio )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zgesdd_US') and sdd:
			plot_gesvd_data( v.zgesdd_US, True, 'x--', color=v.color, label=v.version+' zgesdd', ratio=ratio )
		if v.__dict__.has_key('zgesvd_US') and svd:
			plot_gesvd_data( v.zgesvd_US, True, '.-',  color=v.color, label=v.version+' zgesvd', ratio=ratio )
	# end
	
	# plot lapack last; stop after 1st occurence
	for v in versions[-1::-1]:
		if ( lapack and v.__dict__.has_key('sgesdd_US') and not isnan( v.sgesdd_US[0,svd_cpu_time] )):
			print( 'found LAPACK in', v.version )
			
			# for vectors, gesdd > gesvd performance
			subplot( 2, 2, 1 )
			if v.__dict__.has_key('sgesdd_US') and sdd:
				plot_gesvd_data( v.sgesdd_US, True, 'x--', color='k', label='MKL sgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('sgesvd_US') and svd:
				plot_gesvd_data( v.sgesvd_US, True, '.-',  color='k', label='MKL sgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 2 )
			if v.__dict__.has_key('dgesdd_US') and sdd:
				plot_gesvd_data( v.dgesdd_US, True, 'x--', color='k', label='MKL dgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('dgesvd_US') and svd:
				plot_gesvd_data( v.dgesvd_US, True, '.-',  color='k', label='MKL dgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 3 )
			if v.__dict__.has_key('cgesdd_US') and sdd:
				plot_gesvd_data( v.cgesdd_US, True, 'x--', color='k', label='MKL cgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('cgesvd_US') and svd:
				plot_gesvd_data( v.cgesvd_US, True, '.-',  color='k', label='MKL cgesvd', ratio=ratio, idx=svd_cpu_time )
			
			subplot( 2, 2, 4 )
			if v.__dict__.has_key('zgesdd_US') and sdd:
				plot_gesvd_data( v.zgesdd_US, True, 'x--', color='k', label='MKL zgesdd', ratio=ratio, idx=svd_cpu_time )
			if v.__dict__.has_key('zgesvd_US') and svd:
				plot_gesvd_data( v.zgesvd_US, True, '.-',  color='k', label='MKL zgesvd', ratio=ratio, idx=svd_cpu_time )
			break
	# end
	
	m = ratio if (ratio >= 1) else 1
	n = 1     if (ratio >= 1) else 1/ratio
	subplot( 2, 2, 1 )
	plot_gesvd_labels( versions, 'sgesvd, some vectors, M:N ratio %.3g:%.3g' % (m,n), vec=True, square=(ratio == 1), legend= True )
	
	subplot( 2, 2, 2 )
	plot_gesvd_labels( versions, 'dgesvd, some vectors, M:N ratio %.3g:%.3g' % (m,n), vec=True, square=(ratio == 1), legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_gesvd_labels( versions, 'cgesvd, some vectors, M:N ratio %.3g:%.3g' % (m,n), vec=True, square=(ratio == 1), legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_gesvd_labels( versions, 'zgesvd, some vectors, M:N ratio %.3g:%.3g' % (m,n), vec=True, square=(ratio == 1), legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'gesvd_js.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_symv_data( data, uplo, style='.-', color='y', label=None, first=False ):
	uplo += ' '
	if ( first ):
		if ( g_log ):
			pp.semilogx( data[:,symv_n], data[:,symv_atomics_flops], style, markevery=10, color='#aaaaaa', lw=1.5, label=uplo + 'cublas atomics' )
			pp.semilogx( data[:,symv_n], data[:,symv_cublas_flops],  style, markevery=10, color='#666666', lw=1.5, label=uplo + 'cublas'         )
			pp.semilogx( data[:,symv_n], data[:,symv_cpu_flops],     style, markevery=10, color='black',   lw=1.5, label=uplo + 'MKL'            )
		else:
			pp.plot(     data[:,symv_n], data[:,symv_atomics_flops], style, markevery=10, color='#aaaaaa', lw=1.5, label=uplo + 'cublas atomics' )
			pp.plot(     data[:,symv_n], data[:,symv_cublas_flops],  style, markevery=10, color='#666666', lw=1.5, label=uplo + 'cublas'         )
			pp.plot(     data[:,symv_n], data[:,symv_cpu_flops],     style, markevery=10, color='black',   lw=1.5, label=uplo + 'MKL'            )
	if ( g_log ):
		pp.semilogx( data[:,symv_n], data[:,symv_gpu_flops], style, markevery=10, color=color, lw=1.5, label=uplo + label )
	else:
		pp.plot(     data[:,symv_n], data[:,symv_gpu_flops], style, markevery=10, color=color, lw=1.5, label=uplo + label )
# end

def plot_symv_labels( versions, title=None, legend=True ):
	set_title( versions, 'Symmetric matrix-vector multiply (symv) ' + str_none(title) )
	if legend:
		pp.legend( loc='upper left' )
	pp.ylabel( r'Gflop/s  $2n^2 / t$' )
	set_xticks()
	#(ymin, ymax) = pp.ylim()
	#pp.ylim( 0, min(310, ymax) )
# end

def plot_symv( versions ):
	print( 'plotting symv' )
	versions = get_versions( versions )
	figure( 10 )
	clf( 2, 2 )
	
	first = True
	for v in versions:
		subplot( 2, 2, 1 )
		if v.__dict__.has_key('ssymv_L'):
			plot_symv_data(  v.ssymv_L, 'lower', '-',   color=v.color, label=v.version+' ssymv', first=first )
		if v.__dict__.has_key('ssymv_U'):
			plot_symv_data(  v.ssymv_U, 'upper', 'x--', color=v.color, label=v.version+' ssymv', first=first )
		
		subplot( 2, 2, 2 )
		if v.__dict__.has_key('dsymv_L'):
			plot_symv_data(  v.dsymv_L, 'lower', '-',   color=v.color, label=v.version+' dsymv', first=first )
		if v.__dict__.has_key('dsymv_U'):
			plot_symv_data(  v.dsymv_U, 'upper', 'x--', color=v.color, label=v.version+' dsymv', first=first )
		
		subplot( 2, 2, 3 )
		if v.__dict__.has_key('chemv_L'):
			plot_symv_data(  v.chemv_L, 'lower', '-',   color=v.color, label=v.version+' chemv', first=first )
		if v.__dict__.has_key('chemv_U'):
			plot_symv_data(  v.chemv_U, 'upper', 'x--', color=v.color, label=v.version+' chemv', first=first )
		
		subplot( 2, 2, 4 )
		if v.__dict__.has_key('zhemv_L'):
			plot_symv_data(  v.zhemv_L, 'lower', '-',   color=v.color, label=v.version+' zhemv', first=first )
		if v.__dict__.has_key('zhemv_U'):
			plot_symv_data(  v.zhemv_U, 'upper', 'x--', color=v.color, label=v.version+' zhemv', first=first )
		first = False
	# end
	
	subplot( 2, 2, 1 )
	plot_symv_labels( versions, 'ssymv', legend= True )
	
	subplot( 2, 2, 2 )
	plot_symv_labels( versions, 'dsymv', legend= not g_subplots )
	
	subplot( 2, 2, 3 )
	plot_symv_labels( versions, 'chemv', legend= not g_subplots )
	
	subplot( 2, 2, 4 )
	plot_symv_labels( versions, 'zhemv', legend= not g_subplots )
	
	resize( g_figsize, 2, 2 )
	savefig( 'symv.pdf', 2, 2 )
# end


# ----------------------------------------------------------------------
def plot_all( versions, lapack=True, cpu=True, gpu=True, bulge=True, sdd=True, svd=True, ratio=1 ):
	plot_getrf( versions, lapack=lapack, cpu=cpu, gpu=gpu )
	plot_potrf( versions, lapack=lapack, cpu=cpu, gpu=gpu )
	plot_geqrf( versions, lapack=lapack, cpu=cpu, gpu=gpu )
	plot_geev(  versions, lapack=lapack )
	plot_syev(  versions, lapack=lapack, cpu=cpu, gpu=gpu, bulge=bulge )
	plot_gesvd( versions, lapack=lapack, ratio=ratio, svd=svd, sdd=sdd )
	plot_symv(  versions )
# end


# ----------------------------------------------------------------------
def print_help():
	print('''Global settings:
g_save      # True to save plots as PDF files
g_subplots  # True for all 4 precisions as subplots in one figure, False for 4 separate figures
g_log       # True for semilogx, False for linear plot
g_figsize   # size of figure with 4-up subplots, default (9,7)
g_figsize2  # size of individual figures, default (6,4)

Available plots:
plot_getrf( versions, lapack=True, cpu=True, gpu=True )
plot_potrf( versions, lapack=True, cpu=True, gpu=True )
plot_geqrf( versions, lapack=True, cpu=True, gpu=True )
plot_geev(  versions, lapack=True )
plot_syev(  versions, lapack=True, cpu=True, gpu=True, bulge=True )
plot_gesvd( versions, lapack=True, svd=True, sdd=True, ratio=1 )
		where ratio m:n in { 1, 3, 100, 1/3., 1/100. }
plot_symv(  versions, lapack=True )

plot_all(   versions, lapack=True, cpu=True, gpu=True, bulge=True, sdd=True, svd=True, ratio=1 )

Available versions:''')
	
	for i in xrange( len(versions) ):
		print( "versions[%d] = %s" % (i, versions[i].version) )
	# end
# end


if ( __name__ == '__main__' ):
	print_help()



#!/usr/bin/env python
#
# MAGMA (version 2.6.2) --
# Univ. of Tennessee, Knoxville
# Univ. of California, Berkeley
# Univ. of Colorado, Denver
# @date March 2022

## @file
#  @author Mark Gates
#  @author Hartwig Anzt
#
# Script to run testers with various matrix sizes.
# Small sizes are chosen around block sizes (e.g., 30...34 around 32) to
# detect bugs that occur at the block size, and the switch over from
# LAPACK to MAGMA code.
# Tall and wide sizes are chosen to exercise different aspect ratios,
# e.g., nearly square, 2:1, 10:1, 1:2, 1:10.
# The -h or --help option provides a summary of the options.
#
# Batch vs. interactive mode
# --------------------------
# When output is redirected to a file, it runs in batch mode, printing a
# short summary to stderr on the console and all other output to the file.
# For example:
#
#       ./run_tests.py --lu --precision s --small > lu.txt
#       testing_sgesv_gpu -c                      ok
#       testing_sgetrf_gpu -c2                    ok
#       testing_sgetf2_gpu -c                     ok
#       testing_sgetri_gpu -c                     ** 45 tests failed
#       testing_sgetrf_mgpu -c2                   ok
#       testing_sgesv -c                          ok
#       testing_sgetrf -c2                        ok
#
#       ****************************************************************************************************
#       summary
#       ****************************************************************************************************
#         282 tests in 7 commands passed
#          45 tests failed accuracy test
#           0 errors detected (crashes, CUDA errors, etc.)
#       routines with failures:
#           testing_sgetri_gpu -c
#
# When output is to console (tty), it runs in interactive mode, pausing after
# each test. At the pause, typing "M" re-makes and re-runs that tester,
# while typing enter goes to the next tester.
# For example (some output suppressed with ... for brevity):
#
#       ./run_tests.py --lu --precision s --small
#       ****************************************************************************************************
#       ./testing_sgesv_gpu -c --range 1:20:1 ...
#       ****************************************************************************************************
#           N  NRHS   CPU GFlop/s (sec)   GPU GFlop/s (sec)   ||B - AX|| / N*||A||*||X||
#       ================================================================================
#           1     1     ---   (  ---  )      0.00 (   0.00)   9.26e-08   ok
#           2     1     ---   (  ---  )      0.00 (   0.00)   1.32e-08   ok
#           3     1     ---   (  ---  )      0.00 (   0.00)   8.99e-09   ok
#       ...
#         ok
#       [enter to continue; M to make and re-run]
#
#       ****************************************************************************************************
#       ./testing_sgetri_gpu -c --range 1:20:1 ...
#       ****************************************************************************************************
#       % MAGMA 1.4.0 svn compiled for CUDA capability >= 3.0
#       % CUDA runtime 6000, driver 6000. MAGMA not compiled with OpenMP.
#       % device 0: GeForce GT 750M, 925.5 MHz clock, 2047.6 MB memory, capability 3.0
#       Usage: ./testing_sgetri_gpu [options] [-h|--help]
#
#           N   CPU GFlop/s (sec)   GPU GFlop/s (sec)   ||R||_F / (N*||A||_F)
#       =================================================================
#           1      0.00 (   0.00)      0.00 (   0.00)   6.87e+01   failed
#           2      0.00 (   0.00)      0.00 (   0.00)   2.41e+00   failed
#           3      0.01 (   0.00)      0.00 (   0.00)   1.12e+00   failed
#       ...
#         ** 45 tests failed
#       [enter to continue; M to make and re-run]
#
#       ...
#
#       ****************************************************************************************************
#       summary
#       ****************************************************************************************************
#         282 tests in 7 commands passed
#          45 tests failed accuracy test
#           0 errors detected (crashes, CUDA errors, etc.)
#       routines with failures:
#           testing_sgetri_gpu -c
#
#
# What tests are run
# ------------------
# The --blas, --aux, --chol, --lu, --qr, --syev, --sygv, --geev, --svd options run
# particular sets of tests. By default, all tests are run.
#
# The --start option skips all testers before the given one, then continues
# with testers from there. This is helpful to restart a batch. For example:
#
#       ./run_tests.py --start testing_spotrf
#
# If specific testers are named on the command line, only those are run.
# For example:
#
#       ./run_tests.py testing_spotrf testing_sgetrf
#
# The -p/--precision option controls what precisions are tested, the default
# being "sdcz" for all four precisions. For example, to run single and double:
#
#       ./run_tests.py -p sd
#
# The -s/--small, -m/--medium, -l/--large options control what sizes are tested,
# the default being all three sets.
#       -s/--small  does small  tests, N < 300.
#       -m/--medium does medium tests, N < 1000.
#       -l/--large  does large  tests, N > 1000.
# For example, running small and medium tests:
#
#       ./run_tests.py -s -m
#
#
# What is checked
# ------------------
# The --memcheck option runs cuda-memcheck. This is very helpful for finding
# memory bugs (reading & writing outside allocated memory). It is, however, slow.
#
# The --tol option sets the tolerance to verify accuracy. This is 30 by default,
# which may be too tight for some testers. Setting it somewhat higher
# (e.g., 50 or 100) filters out spurious accuracy failures.

import os
import re
import sys
import time

import subprocess
from subprocess import PIPE, STDOUT

from optparse import OptionParser

# on a TTY screen, stop after each test for user input
# when redirected to file ("batch mode"), don't stop
batch = not sys.stdout.isatty()

parser = OptionParser()
parser.add_option('-p', '--precisions',  action='store',      dest='precisions', help='run given precisions', default='sdcz' )
parser.add_option(      '--start',       action='store',      dest='start',      help='start with given routine; useful to restart an interupted run')
parser.add_option(      '--memcheck',    action='store_true', dest='memcheck',   help='run with cuda-memcheck (slow)')
parser.add_option(      '--tol',         action='store',      dest='tol',        help='set tolerance')
                                         
parser.add_option('-s', '--small',       action='store_true', dest='small',      help='run small  tests, N < 300')
parser.add_option('-m', '--medium',      action='store_true', dest='med',        help='run medium tests, N < 1000')
parser.add_option('-l', '--large',       action='store_true', dest='large',      help='run large  tests, N > 1000')
parser.add_option('-n', '--nonsym',      action='store_true', dest='nonsym',     help='run nonsym  tests')

parser.add_option(      '--sparse-blas', action='store_true', dest='sparse_blas', help='run sparse BLAS tests')
parser.add_option(      '--solver',      action='store_true', dest='solver',      help='run sparse solvers')
parser.add_option(      '--control',     action='store_true', dest='control',     help='run sparse IO, copy, etc.')

parser.add_option(      '--csr',         action='store_true', dest='csr',         help='run CSR matrix format')
parser.add_option(      '--ell',         action='store_true', dest='ell',         help='run ELL matrix format')
parser.add_option(      '--sellp',       action='store_true', dest='sellp',       help='run SELLP matrix format')
parser.add_option(      '--csr5',        action='store_true', dest='csr5',        help='run CSR5 matrix format')

parser.add_option(      '--cg'               , action='store_true', dest='cg'            , help='run cg'            )
parser.add_option(      '--cg_merge'         , action='store_true', dest='cg_merge'      , help='run cg_merge'      )
parser.add_option(      '--pcg'              , action='store_true', dest='pcg'           , help='run pcg'           )
parser.add_option(      '--bicgstab'         , action='store_true', dest='bicgstab'      , help='run bicgstab'      )
parser.add_option(      '--bicgstab_merge'   , action='store_true', dest='bicgstab_merge', help='run bicgstab_merge')
parser.add_option(      '--pbicgstab'        , action='store_true', dest='pbicgstab'     , help='run pbicgstab'     )
parser.add_option(      '--gmres'            , action='store_true', dest='gmres'         , help='run gmres'         )
parser.add_option(      '--pgmres'           , action='store_true', dest='pgmres'        , help='run pgmres'        )
parser.add_option(      '--lobpcg'           , action='store_true', dest='lobpcg'        , help='run lobpcg'        )
parser.add_option(      '--iterref'          , action='store_true', dest='iterref'       , help='run iterref'       )
parser.add_option(      '--jacobi'           , action='store_true', dest='jacobi'        , help='run jacobi'        )
parser.add_option(      '--ba'               , action='store_true', dest='ba'            , help='run ba-iter'       )
parser.add_option(      '--idr'              , action='store_true', dest='idr'           , help='run idr'           )
parser.add_option(      '--idr_merge'        , action='store_true', dest='idr_merge'     , help='run idr_merge'     )
parser.add_option(      '--pidr'             , action='store_true', dest='pidr'          , help='run pidr'          )
parser.add_option(      '--cgs'              , action='store_true', dest='cgs'           , help='run cgs'           )
parser.add_option(      '--cgs_merge'        , action='store_true', dest='cgs_merge'     , help='run cgs_merge'     )
parser.add_option(      '--qmr'              , action='store_true', dest='qmr'           , help='run qmr'           )
parser.add_option(      '--qmr_merge'        , action='store_true', dest='qmr_merge'     , help='run qmr_merge'     )
parser.add_option(      '--pcgs'             , action='store_true', dest='pcgs'          , help='run pcgs'          )
parser.add_option(      '--pcgs_merge'       , action='store_true', dest='pcgs_merge'    , help='run pcgs_merge'    )
parser.add_option(      '--pqmr'             , action='store_true', dest='pqmr'          , help='run pqmr'          )
parser.add_option(      '--pqmr_merge'       , action='store_true', dest='pqmr_merge'    , help='run pqmr_merge'    )   
parser.add_option(      '--bombard'          , action='store_true', dest='bombard'       , help='run bombard'       )
parser.add_option(      '--bombard_merge'    , action='store_true', dest='bombard_merge' , help='run bombard_merge' )
parser.add_option(      '--lsqr'             , action='store_true', dest='lsqr'          , help='run lsqr'          )
parser.add_option(      '--bicg'             , action='store_true', dest='bicg'          , help='run bicg'          )
parser.add_option(      '--pbicg'            , action='store_true', dest='pbicg'         , help='run pbicg'         )

                                                                                           
parser.add_option(      '--jacobi-prec'      , action='store_true', dest='jacobi_prec'   , help='run Jacobi preconditioner')
parser.add_option(      '--ilu-prec'         , action='store_true', dest='ilu_exact_prec', help='run ILU + exact solve preconditioner')
parser.add_option(      '--ilu-jac'          , action='store_true', dest='ilu_jac_prec',   help='run ILU + Jacobi solve preconditioner')
parser.add_option(      '--ilu-bjac'         , action='store_true', dest='ilu_bjac_prec',  help='run ILU + Block Jacobi solve preconditioner')
parser.add_option(      '--ilu-isai-prec'    , action='store_true', dest='ilu_isai_prec' , help='run ILU + ISAI preconditioner')
parser.add_option(      '--ilut-prec'        , action='store_true', dest='ilut_prec',      help='run threshold ILU + exact solve preconditioner')

(opts, args) = parser.parse_args()

# default if no sizes given is all sizes
if ( not opts.small and not opts.med and not opts.large and not opts.nonsym ):
    opts.small  = True
    opts.med    = False
    opts.large  = False
    opts.nonsym = True
# end

# default if no groups given is all groups
if (     not opts.sparse_blas
     and not opts.solver
     and not opts.control
     and not opts.csr
     and not opts.ell
     and not opts.sellp ):
    opts.sparse_blas = True
    opts.solver      = True
    opts.control     = True
    opts.csr         = True
    opts.ell         = True
    opts.sellp       = True
    opts.csr5        = True
# end

# default if no solvers given is all solvers
if (     not opts.cg
     and not opts.cg_merge
     and not opts.pcg
     and not opts.bicgstab
     and not opts.bicgstab_merge
     and not opts.pbicgstab
     and not opts.gmres
     and not opts.pgmres
     and not opts.lobpcg
     and not opts.iterref
     and not opts.jacobi
     and not opts.ba
     and not opts.idr
     and not opts.ba
     and not opts.idr_merge
     and not opts.qmr
     and not opts.qmr_merge
     and not opts.cgs
     and not opts.cgs_merge
     and not opts.pqmr
     and not opts.pqmr_merge
     and not opts.pcgs
     and not opts.pcgs_merge
     and not opts.bombard
     and not opts.bombard_merge
     and not opts.bicg
     and not opts.pbicg
     and not opts.lsqr
     and not opts.pidr ):
    opts.cg             = True
    opts.cg_merge       = True
    opts.pcg            = True
    opts.bicgstab       = True
    opts.bicgstab_merge = True
    opts.pbicgstab      = True
    opts.gmres          = True
    opts.pgmres         = True
    opts.lobpcg         = True
    opts.iterref        = True
    opts.jacobi         = True
    opts.ba             = True
    opts.idr            = True
    opts.idr_merge      = True
    opts.pidr           = True
    opts.cgs            = True
    opts.cgs_merge      = True
    opts.qmr            = True
    opts.qmr_merge      = True
    opts.pcgs           = True
    opts.pcgs_merge     = True
    opts.pqmr           = True
    opts.pqmr_merge     = True
    opts.bombard        = True
    opts.bombard_merge  = True
    opts.bicg           = True
    opts.pbicg          = True
    opts.lsqr           = True
# end

# default if no preconditioners given all
if (     not opts.jacobi_prec
     and not opts.ilu_exact_prec 
     and not opts.ilut_prec
     and not opts.ilu_jac_prec
     and not opts.ilu_bjac_prec
     and not opts.ilu_isai_prec ):
    opts.jacobi_prec      = True
    opts.ilu_prec         = True
    opts.ilu_jac_prec     = True
    opts.ilu_isai_prec    = True
    opts.ilu_bjac_prec    = True
    opts.ilut_prec        = True
# end

# default if no sizes given is all sizes
if ( not opts.small and not opts.med and not opts.large and not opts.nonsym ):
    opts.small = True
    opts.med   = True
    opts.large = True
    opts.nonsym = True
# end

print 'opts', opts
print 'args', args


# ----------------------------------------------------------------------
    
# looping over formats
formats = []
if ( opts.csr ):
    formats += ['--format CSR']
# end
if ( opts.ell ):
    formats += ['--format ELL']
# end
if ( opts.sellp ):
    formats += ['--format SELLP']
# end
if ( opts.csr5 ):
    formats += ['--format CSR5']
# end

# looping over solvers
solvers = []
if ( opts.cg ):
    solvers += ['--solver CG --basic']
# end
if ( opts.cg_merge ):
    solvers += ['--solver CG']
# end
if ( opts.bicgstab ):
    solvers += ['--solver BICGSTAB --basic']
# end
if ( opts.bicgstab_merge ):
    solvers += ['--solver BICGSTAB']
# end
if ( opts.gmres ):
    solvers += ['--solver GMRES']
# end
if ( opts.lobpcg ):
    solvers += ['--solver LOBPCG']
# end
if ( opts.jacobi ):
    solvers += ['--solver JACOBI']
# end
if ( opts.ba ):
    solvers += ['--solver BA']
# end
if ( opts.idr ):
    solvers += ['--solver IDR --basic']
# end
if ( opts.idr_merge ):
    solvers += ['--solver IDR']
# end
if ( opts.cgs ):
    solvers += ['--solver CGS --basic']
# end
if ( opts.cgs_merge ):
    solvers += ['--solver CGS --basic']
# end
if ( opts.qmr ):
    solvers += ['--solver QMR --basic']
# end
if ( opts.qmr_merge ):
    solvers += ['--solver QMR']
# end
if ( opts.lsqr ):
    solvers += ['--solver LSQR']
# end
if ( opts.bicg ):
    solvers += ['--solver BICG']
# end
if ( opts.bombard ):
    solvers += ['--solver BOMBARDMENT']
# end
if ( opts.bombard_merge ):
    solvers += ['--solver BOMBARDMENT --basic']
# end


# looping over precsolvers
precsolvers = []
if ( opts.pcg ):
    precsolvers += ['--solver PCG ']
# end
if ( opts.pbicgstab ):
    precsolvers += ['--solver PBICGSTAB ']
# end
if ( opts.pgmres ):
    precsolvers += ['--solver PGMRES ']
# end
if ( opts.pidr ):
    precsolvers += ['--solver PIDR ']
# end
if ( opts.pqmr ):
    precsolvers += ['--solver PQMR ']
# end
if ( opts.pidr ):
    precsolvers += ['--solver PCGS ']
# end
if ( opts.pbicg ):
    precsolvers += ['--solver PBICG ']
# end
if ( opts.lsqr ):
    precsolvers += ['--solver PLSQR ']
# end



# looping over eigensolvers
IR = []
if ( opts.iterref ):
    IR += ['--solver LOBPCG ']
# end



# looping over preconditioners
precs = ['--precond NONE ']
if ( opts.jacobi_prec ):
    precs += ['--precond JACOBI ']
# end
if ( opts.ilu_exact_prec ):
    precs += ['--precond ILU ']
# end
if ( opts.ilut_prec ):
    precs += ['--precond PARILUT --prestart 1 --psweeps 5 --plevels 0 --prtol 0.05 --patol 0.2 ']
# end
if ( opts.ilu_jac_prec ):
    precs += ['--precond ILU --trisolver JACOBI --ppattern 1 --piters 5 ']
# end
if ( opts.ilu_bjac_prec ):
    precs += ['--precond ILU --trisolver JACOBI --ppattern 4 --piters 3 ']
# end
if ( opts.ilu_isai_prec ):
    precs += ['--precond ILU --trisolver ISAI --ppattern 1 --piters 1 ']
# end


# looping over preconditioners for Iter-Ref
IRprecs = []
if ( opts.iterref ):
    IRprecs += ['--precond CG ']
    IRprecs += ['--precond CGS ']
    IRprecs += ['--precond BICGSTAB ']
    IRprecs += ['--precond BICG ']
    IRprecs += ['--precond GMRES ']
    IRprecs += ['--precond QMR ']
# end



# ----------------------------------------------------------------------
# blocksizes
# ----------
blocksizes = []
blocksizes += ['--blocksize 4']
blocksizes += ['--blocksize 8']
blocksizes += ['--blocksize 16']
#end


# ----------------------------------------------------------------------
# alignments
# ----------
alignments = []
alignments += ['--alignment 4']
alignments += ['--alignment 8']
alignments += ['--alignment 16']
alignments += ['--alignment 32']


# ----------------------------------------------------------------------
# problem sizes
# n    is squared
# ----------
sizes = []
if opts.small:
    sizes += ['LAPLACE2D 47']
if opts.med:
    sizes += ['LAPLACE2D 95']
if opts.large:
    sizes += ['LAPLACE2D 317']
if opts.nonsym:
    sizes += ['test_matrices/ani5_crop.mtx']
#end


# ----------------------------------------------------------------------
precisions = (
    's', 'd', 'c', 'z'
)

subs = (
    ('',              'dlag2s',      '',              'zlag2c'    ),
    ('ssy',           'dsy',         'che',           'zhe'       ),
    ('sor',           'dor',         'cun',           'zun'       ),
    ('sy2sb',         'sy2sb',       'he2hb',         'he2hb'     ),
    ('',              'testing_ds',  '',              'testing_zc'),
    ('testing_s',     'testing_d',   'testing_c',     'testing_z' ),
    ('lansy',         'lansy',       'lanhe',         'lanhe'     ),
    ('blas_s',        'blas_d',      'blas_c',        'blas_z'    ),
)

# ----------
# simple precision generation
def substitute( txt, pfrom, pto ):
    if ( pfrom != pto ):
        ifrom = precisions.index( pfrom )
        ito   = precisions.index( pto )
        for sub in subs:
            txt = re.sub( sub[ifrom], sub[ito], txt )
    # end
    return txt
# end


# ----------------------------------------------------------------------
# test problems
#
# Each test has 4 fields:  (program, options, sizes, comments).
# If program begins with #, that test is disabled and will be printed as such.
# In some cases, the line is commented out, so nothing is printed.
#
# These match the order in the Makefile,
# except in some cases the "d" version from the Makefile isn't required here
# To compare with Makefile:
# grep '^\s+testing_\w+' Makefile >! a
# grep "\('#?testing_\w\w+" run_tests.py | perl -pe "s/#?\('#?//; s/',.*/.cpp \\/;" | uniq > ! b
# diff -w a b


tests = []


# ----------------------------------------------------------------------
if ( opts.control):
    for precision in opts.precisions:
        for size in sizes:
            # precision generation
            cmd = substitute( 'testing_zio', 'z', precision )
            tests.append( [cmd, '', size, ''] )

# ----------------------------------------------------------------------
if ( opts.control):
    for precision in opts.precisions:
        for size in sizes:
            # precision generation
            cmd = substitute( 'testing_zmatrix', 'z', precision )
            tests.append( [cmd, '', size, ''] )

# ----------------------------------------------------------------------
if ( opts.control):
    for precision in opts.precisions:
        for size in sizes:
            # precision generation
            cmd = substitute( 'testing_zmconverter', 'z', precision )
            tests.append( [cmd, '', size, ''] )

# ----------------------------------------------------------------------
if ( opts.control):
    for precision in opts.precisions:
        for size in sizes:
            # precision generation
            cmd = substitute( 'testing_zmcompressor', 'z', precision )
            tests.append( [cmd, '', size, ''] )

# ----------------------------------------------------------------------
if ( opts.control):
    for precision in opts.precisions:
        for size in sizes:
            # precision generation
            cmd = substitute( 'testing_zmadd', 'z', precision )
            tests.append( [cmd, '', size + ' ' + size, ''] )


# ----------------------------------------------------------------------
if ( opts.sparse_blas):
    for precision in opts.precisions:
        for size in sizes:
            for blocksize in blocksizes:
                for alignment in alignments:
                    # precision generation
                    cmd = substitute( 'testing_zspmv', 'z', precision )
                    tests.append( [cmd, alignment + ' ' + blocksize, size, ''] )


# ----------------------------------------------------------------------
if ( opts.sparse_blas):
    for precision in opts.precisions:
        for size in sizes:
            for blocksize in blocksizes:
                for alignment in alignments:
                    # precision generation
                    cmd = substitute( 'testing_zspmm', 'z', precision )
                    tests.append( [cmd, alignment + ' ' + blocksize, size, ''] )


# ----------------------------------------------------------------------
for solver in solvers:
    for size in sizes:
        for precision in opts.precisions:
            # precision generation
            cmd = substitute( 'testing_zsolver', 'z', precision )
            tests.append( [cmd, solver, size, ''] )


# ----------------------------------------------------------------------
for solver in precsolvers:
    for precond in precs:
        for size in sizes:
            for precision in opts.precisions:
                # precision generation
                cmd = substitute( 'testing_zsolver', 'z', precision )
                tests.append( [cmd, solver + ' ' + precond, size, ''] )


# ----------------------------------------------------------------------
for solver in IR:
    for precond in IRprecs:
        for size in sizes:
            for precision in opts.precisions:
                # precision generation
                cmd = substitute( 'testing_zsolver', 'z', precision )
                tests.append( [cmd, solver + ' ' + precond, size, ''] )








# ----------------------------------------------------------------------
#print 'tests'
#for t in tests:
#    print t


# ----------------------------------------------------------------------
# runs command in a subprocess.
# returns list (okay, fail, errors, status)
# okay   is count of "ok"     in output.
# fail   is count of "failed" in output.
# error  is count of indications of other errors (exit, CUDA error, etc.).
# status is exit status of the command.
def run( cmd ):
    words = re.split( ' +', cmd.strip() )
    
    # stdout & stderr are merged
    p = subprocess.Popen( words, bufsize=1, stdout=PIPE, stderr=STDOUT )
    
    okay  = 0
    fail  = 0
    nospd = 0
    error = 0
    nosupport = 0
    slowconv = 0
    # read unbuffered ("for line in p.stdout" will buffer)
    while True:
        line = p.stdout.readline()
        if not line:
            break
        print line.rstrip()
        if re.search( r'\bok *$', line ):
            okay += 1
        if re.search( 'failed', line ):
            fail += 1
        if re.search( 'exit|memory mapping error|CUDA runtime error|illegal value|ERROR SUMMARY: [1-9]', line ):
            error += 1
        m = re.search( 'solver info: (-?\d+)', line )
        if ( m ):
            info = int( m.group(1) )
            if ( info == 0 ):
                okay += 1
            if ( info == -103 ):
                nosupport += 1
            if ( info == -201 ):
                slowconv += 1
            if ( info == -202 ):
                fail += 1
            if ( info == -203 ):
                fail += 1
        # end
    # end
    
    status = p.wait()
    return (okay, fail, error, status)
# end


# ----------------------------------------------------------------------
ntest  = 0
nokay  = 0
nfail  = 0
nerror = 0
failures = {}

start = None
if ( opts.start ):
    start = re.compile( opts.start + r'\b' )

seen  = {}
pause = 0

last_cmd = None

for test in tests:
    (cmd, options, sizes, comments) = test
    
    make = False
    disabled = (cmd[0] == '#')
    if ( disabled ):
        cmd = cmd[1:]
    
    # command to run
    cmd_args = './' + cmd +' '+ options +' '+ sizes
    cmd_args = re.sub( '  +', ' ', cmd_args )  # compress spaces
    
    # skip tests before start
    if ( start and not start.search( cmd_args )):
        continue
    start = None
    
    # skip tests not in args, or duplicates, or non-existing
    #if not os.path.exists( cmd ):
    #    print >>sys.stderr, cmd, cmd, "doesn't exist"
    if (    (args and not cmd in args)
         or (not os.path.exists( cmd ))
         or (seen.has_key( cmd_args )) ):
        #print "skipping", cmd_args
        continue
    # end
    seen[ cmd_args ] = True
    
    if ( opts.memcheck ):
        cmd_args = 'cuda-memcheck ' + cmd_args
    
    go = True
    while( go ):
        if pause > 0:
            time.sleep( pause )
            pause = 0
        # end
        
        print
        print '*'*100
        print cmd_args
        print '*'*100
        sys.stdout.flush()
        
        if ( batch ):
            if ( last_cmd and cmd != last_cmd ):
                sys.stderr.write( '\n' )
            last_cmd = cmd
            sys.stderr.write( '%-70s' % cmd_args )
            sys.stderr.flush()
        # end
        
        if ( disabled ):
            if ( comments ):
                sys.stderr.write( '  (disabled: ' + comments + ')\n' )
            else:
                sys.stderr.write( '  (disabled)\n' )
            go = False
            continue
        # end
        
        if ( make ):
            m = 'make lib ' + cmd
            print m
            run( m )
        # end
        
        t = time.time()
        (okay, fail, error, status) = run( cmd_args )
        t = time.time() - t
        
        # count stats
        ntest  += 1
        nokay  += okay
        nfail  += fail
        nerror += error
        
        errmsg = ''
        if ( fail > 0 ):
            errmsg += '  ** %d tests failed' % (fail)
        if ( error > 0 ):
            errmsg += '  ** %d errors' % (error)
        if ( status < 0 ):
            errmsg += '  ** exit with signal %d' % (-status)
            nerror += 1  # count crash as an error
        
        if ( errmsg != '' ):
            if ( batch ):
                sys.stderr.write( errmsg + '\n' )  # to console
            sys.stdout.write( errmsg + '\n' )  # to file
            failures[ cmd_args ] = True
        else:
            sys.stderr.write( '  ok\n' )
        # end
        
        if ( batch ):
            # set to sleep a few seconds before next test,
            # to allow processor to cool off some between tests.
            pause = min( t, 5. )
            go = False
        else:
            x = raw_input( '[enter to continue; M to make and re-run] ' )
            if ( x in ('m','M')):
                make = True
            else:
                go = False
        # endif
    # end
# end


# print summary
msg  = '\n'
msg += '*'*100   + '\n'
msg += 'summary' + '\n'
msg += '*'*100   + '\n'

if ( nfail == 0 and nerror == 0 ):
    msg += 'all %d tests in %d commands passed!\n' % (nokay, ntest)
else:
    msg += '%5d tests in %d commands passed\n' % (nokay, ntest)
    msg += '%5d tests failed accuracy test\n' % (nfail)
    msg += '%5d errors detected (crashes, CUDA errors, etc.)\n' % (nerror)
    f = failures.keys()
    f.sort()
    msg += 'routines with failures:\n    ' + '\n    '.join( f ) + '\n'
# end

if ( batch ):
    sys.stderr.write( msg )  # to console
sys.stdout.write( msg )  # to file



import re
import numpy


# ------------------------------------------------------------
def intersect( A, B ):
	'''
	Finds common elements in A and B
	Returns pair of index arrays (ii, jj) such that a[ii] == b[jj].
	'''
	(ii,) = numpy.where( numpy.in1d( A, B ))
	(jj,) = numpy.where( numpy.in1d( B, A ))
	return (ii, jj)
# end


# ------------------------------------------------------------
def groupby( data, index ):
	'''
	Returns array with unique values in column index,
	and min-reduction in all other columns, like an SQL groupby query.
	'''
	if ( data is None ):
		return None
	
	values = numpy.unique( data[:,index] )
	rows = len(values)
	cols = data.shape[1]
	data2 = numpy.zeros(( rows, cols ))
	for i in xrange( rows ):
		(ii,) = numpy.where( data[:,index] == values[i] )
		for j in xrange( cols ):
			data2[i,j] = numpy.min( data[ii,j] )
	# end
	return data2
# end


# ------------------------------------------------------------
def legend( lines=None, labels=None, loc="best", **kwargs ):
	'''
	Wrapper around pyplot's legend
	Adds "[upper,lower] outside" and "outside [left,right]" as locations.
	
	Adds left, right, upper, lower as aliases to
	to   center left, center right, upper center, lower center, respectively.
	
	Accepts locations with word order transposed,
	e.g., "center lower" instead of "lower center".
	'''
	import matplotlib.pyplot as pp
	
	anchor = None
	locs = re.split( ' +', loc )
	if ( len(locs) == 1 ):
		if   ( locs[0] in ('upper', 'top'   )): locs = ['upper center']
		elif ( locs[0] in ('lower', 'bottom')): locs = ['lower center']
		elif ( locs[0] in ('left'           )): locs = ['center left']
		elif ( locs[0] not in ('right', 'best', 'center')):
			print 'loc=\'' + loc + '\' is invalid; should be one of:'
			print 'best,'
			print '                           upper  outside'
			print '              upper  left, upper  center (upper), upper  right,'
			print 'outside left, center left,        center,         center right, outside right'
			print '                (left)                              (right)'
			print '              lower  left, lower  center (lower), lower  right,'
			print '                           lower  outside'
			print '(top=upper and bottom=lower are recognized synonyms)'
			locs = ['best']
	elif ( len(locs) == 2 ):
		if ( locs[0] in ('left', 'right') or locs[1] in ('lower', 'upper', 'top', 'bottom')):
			locs = [locs[1], locs[0]]  # swap
		if ( locs[0] == 'top'    ): locs[0] = 'upper'
		if ( locs[0] == 'bottom' ): locs[0] = 'lower'
	else:
		raise Exception()
	
	loc2 = ' '.join( locs )
	if ( loc != loc2 ):
		print 'using loc=\'' + loc2 + '\''
	
	if ( loc2 == 'outside right' ):
		anchor = (1, 0.5)
		loc2 = 'center left'
	elif ( loc2 == 'outside left' ):
		anchor = (-0.05, 0.5)  # depends on figure size, which is horrible
		loc2 = 'center right'
	elif ( loc2 == 'upper outside' ):
		anchor = (0.5, 1)
		loc2 = 'lower center'
	elif ( loc2 == 'lower outside' ):
		anchor = (0.5, -0.2)  # depends on figure size, which is horrible
		loc2 = 'upper center'
	# end
	#print 'loc "' + loc2 + '", anchor', anchor
	
	# what's the best way to pass these things? this seems overly cumbersome
	if ( lines and labels ):
		pp.legend( lines, labels, loc=loc2, bbox_to_anchor=anchor, **kwargs )
	elif ( lines ):
		pp.legend( lines, loc=loc2, bbox_to_anchor=anchor, **kwargs )
	else:
		pp.legend( loc=loc2, bbox_to_anchor=anchor, **kwargs )
# end


# ------------------------------------------------------------
def resize( size=None, borders=None, space=None, pad=0.5, extra=None, figures=None ):
	'''
	Resizes and adjusts margins of matplotlib figures.
	
	size    is [ width, height ]
	
	borders is [ top, right, bottom, left ], as in HTML CSS
	
	space   is [ wspace, hspace ]
	
	If borders and space are not given, uses tight_layout()
	pad     is padding used in tight_layout()

	extra   is [ top, right, bottom, left ], extra space to add after tight_layout()
	
	figures is figure object, number, or list of numbers.
	        If None, the current figure (gcf) is used.
	'''
	import matplotlib.pyplot as pp
	
	if ( figures is None ):
		figures = [ pp.gcf() ]
	elif ( type(figures) in (int,numpy.int32) ):
		figures = [ pp.figure( figures ) ]
	elif ( type(figures) in (tuple,list) ):
		figures = map( pp.figure, figures )
	
	for fig in figures:
		if ( borders ):
			fig.subplots_adjust(
				top    = (1-borders[0]),
				right  = (1-borders[1]),
				bottom =    borders[2],
				left   =    borders[3],
			)
		if ( space ):
			fig.subplots_adjust(
				wspace = space[0],
				hspace = space[1]
			)
		if ( size ):
			fig.set_size_inches( size, forward=True )
		if ( not borders and not space ):
			fig.canvas.draw()
			fig.tight_layout( pad=pad )
			fig.canvas.draw()
		# end
		if ( extra ):
			t = fig.subplotpars.top
			l = fig.subplotpars.left
			b = fig.subplotpars.bottom
			r = fig.subplotpars.right
			fig.subplots_adjust(
				top    = t - extra[0],
				right  = r - extra[1],
				bottom = b + extra[2],
				left   = l + extra[3],
			)
			fig.canvas.draw()
	# end
	#pp.draw()
# end resize
